# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import click


class ControllerNeedeGroup(click.Group):
    def command(self, *args, **kwargs):
        def decorator(f):
            f = click.option(
                "--controller-host",
                "-ch",
                required=True,
                help="The host of the controller",
            )(f)
            f = click.option(
                "--controller-port",
                "-cp",
                required=True,
                help="The port of the controller",
            )(f)
            return super(ControllerNeedeGroup, self).command(*args, **kwargs)(f)

        return decorator
