# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


COSMOS_API_PANEL_SUFFIX = "/panel"
COSMOS_API_STATUS_SUFFIX = "/api/status"
COSMOS_API_META_SUFFIX = "/api/meta"
COSMOS_API_REGISTER_SUFFIX = "/api/register"
COSMOS_API_SET_PROFILE_SUFFIX = "/api/set_profile"
COSMOS_API_SET_TRACE_PATH_SUFFIX = "/api/set_trace_path"
COSMOS_API_UNREGISTER_SUFFIX = "/api/unregister"
COSMOS_API_HEARTBEAT_SUFFIX = "/api/heartbeat"
COSMOS_API_NCCL_COMM_INITIATOR_SUFFIX = "/api/nccl/comm_initiator"
COSMOS_API_NCCL_COMM_ACCEPTOR_SUFFIX = "/api/nccl/comm_acceptor"
COSMOS_API_NCCL_COMM_GET_ALL_SUFFIX = "/api/nccl/comm_all"
COSMOS_API_NCCL_COMM_ERROR_SUFFIX = "/api/nccl/comm_error"
COSMOS_API_NCCL_COMM_STORE_CLEAR_SUFFIX = "/api/nccl/comm_store_clear"
COSMOS_API_NEXT_PROMPT_SUFFIX = "/api/next_prompt"
COSMOS_API_ROLLOUT_SUFFIX = "/api/rollout"
COSMOS_API_VALIDATION_REPORT_SUFFIX = "/api/validation_report"
COSMOS_API_POLICY_TRAIN_ACK_SUFFIX = "/api/policy/train_ack"
COSMOS_API_POLICY_SHARD_INFOS_SUFFIX = "/api/policy/shard_infos"
COSMOS_API_ROLLOUT_SHARD_INFOS_SUFFIX = "/api/rollout/shard_infos"
COSMOS_API_POLICY_SHARD_SEND_INSTS_SUFFIX = "/api/policy_shard_send_insts"
COSMOS_API_ROLLOUT_SHARD_RECV_INSTS_SUFFIX = "/api/shard_recv_insts"
COSMOS_API_GET_TRAINABLE_PARAMS_SUFFIX = "/api/get_trainable_params"
