# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import abc
from typing import Optional


"""
Customized for trtllm backend interaction with main process.
"""


class InternalInstruction(abc.ABC):
    pass


class ShutdownInstruction(InternalInstruction):
    pass


class ValidationInstruction(InternalInstruction):
    def __init__(
        self, validation_step: Optional[int] = None, total_steps: Optional[int] = None
    ):
        self.validation_step = validation_step
        self.total_steps = total_steps
